# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
import numpy as np
from torch.nn import functional as F
from torch.autograd import Variable
from ScoreF import score_dot, score_concat, score_generate, local_m, local_p, local_score
class Loung_NMT(nn.Module):
    def __init__(self,input_feature_size,add_feature_size, output_size,
                 lstm_size, score_f, attention_c, feed_input, window_size=10, reverse=True):
        super(Loung_NMT,self).__init__()
        self.score_f = score_f
        self.attention_c = attention_c
        self.window_size = window_size
        self.feed_input = feed_input
        self.reverse = reverse

        self.encoder = nn.LSTM(input_size=input_feature_size,hidden_size=lstm_size,num_layers=1,
                               batch_first=True)
        if not feed_input:
            self.decoder = nn.LSTM(input_size=add_feature_size, hidden_size=lstm_size, num_layers=1,
                                   batch_first=True)
        else:
            self.decoder = nn.LSTM(input_size=add_feature_size+lstm_size,hidden_size=lstm_size,num_layers=1,
                                   batch_first=True)
        self.class_fc_1 = nn.Linear(lstm_size+lstm_size, lstm_size) # 分类全连接层1
        self.class_fc_2 = nn.Linear(lstm_size, output_size) # 分类全连接层2

        self.attention_fc_1 = nn.Linear(lstm_size, lstm_size)
        self.attention_fc_2 = nn.Linear(2*lstm_size, lstm_size)
        self.attention_fc_3 = nn.Linear(lstm_size, 1)

        self.local_fc_1 = nn.Linear(lstm_size, lstm_size)
        self.local_fc_2 = nn.Linear(lstm_size, 1)

        self.final_fc_1 = nn.Linear(2 * lstm_size + add_feature_size, lstm_size)
        self.final_fc_2 = nn.Linear(lstm_size, output_size)

        self.quer_build = nn.Linear(lstm_size + add_feature_size, lstm_size)

    def attention_forward(self,input_embedding, feed_input_h, dec_prev_hidden, enc_output):
        if not self.feed_input:
            dec_lstm_input = input_embedding
        else:
            dec_lstm_input = torch.cat([input_embedding, feed_input_h], dim=2)
        dec_output, dec_hidden  = self.decoder(dec_lstm_input, dec_prev_hidden)

        query = self.quer_build(dec_lstm_input.squeeze())


        if self.score_f == "dot":
            attention_weights = score_dot(enc_output, dec_hidden[0].squeeze(), self.attention_fc_1, self.attention_fc_2, self.attention_fc_3)
        elif self.score_f == "general":
            attention_weights = score_generate(enc_output, dec_hidden[0].squeeze(), self.attention_fc_1, self.attention_fc_2, self.attention_fc_3)
        elif self.score_f == "concat":
            attention_weights = score_concat(enc_output, query, self.attention_fc_1, self.attention_fc_2, self.attention_fc_3)
        else:
            print ("Attention score function input error!")
            exit()
        if self.attention_c == "local_m":
            if self.reverse:
                t = enc_output.size()[1]-1-t
            pt = local_m(dec_hidden[0].squeeze(), t)
            attention_weights = local_score(attention_weights, pt, enc_output.size()[1], self.window_size/2)
        elif self.attention_c == "local_p":
            pt = local_p(dec_hidden[0].squeeze(), self.local_fc_1, self.local_fc_2, enc_output.size()[1])
            attention_weights = local_score(attention_weights, pt, enc_output.size()[1], self.window_size / 2)
        elif self.attention_c == "global":
            pass
        else:
            print ("Attention class input error!")
            exit()
        atten_output = torch.sum(attention_weights * enc_output, dim=1).unsqueeze(1)
        return atten_output,dec_output,dec_hidden, attention_weights
    def forward(self, source_data,target_data, mode = "train",is_gpu=True):
        enc_output, enc_hidden = self.encoder(source_data)

        dec_prev_hidden = [enc_hidden[0],enc_hidden[1]]
            # dec_prev_hidden[0]: 1*bs*lstm_size, dec_prev_hidden[1]: 1*bs*lstm_size
            # dec_h: bs*lstm_size
        feed_input_h = enc_hidden[0].squeeze(0).unsqueeze(1)

            #input_embedding = target_data_embedding[:,i,:].unsqueeze(1)
        atten_output, dec_output, dec_hidden, attention_weights = self.attention_forward(target_data,
                                                                              feed_input_h,
                                                                              dec_prev_hidden,
                                                                               enc_output,)

        out = self.final_fc_2(F.leaky_relu(self.final_fc_1(torch.cat([atten_output,dec_output, target_data],dim=2))))

        return out.reshape(-1,1), attention_weights

if __name__=="__main__":
    deep_nmt = Loung_NMT(input_feature_size=15,add_feature_size=4,output_size=1,
                         lstm_size=256, score_f="concat", attention_c="local_p", feed_input=True, window_size=10,
                         reverse=True).cuda()
    source_data = torch.Tensor(np.zeros([1, 24, 15])).float().cuda()
    target_data = torch.Tensor(np.zeros([1, 1, 4])).float().cuda()
    preds = deep_nmt(source_data, target_data, is_gpu=True)
    print(preds.shape)
