import torch

#from baysen_lstm import Att_NMT
import torch.nn as nn
import torch.optim as optim
from data_loader import TrainData, TestData
from torch.utils.data import DataLoader
import numpy as np

from model import GAT

import argparse




def test(model, genTestData, MAPE, device):
    model.eval()
    # 开始测试
    pred = []
    tagt = []
    with torch.no_grad():
        for i, (en, de, tg) in enumerate(genTestData):

            change_signal = len(pred)
            if change_signal % 1 == 0:
                en_in = en.transpose(-2, -1)

                out, adp = model(en_in.to(device), idx=None, add_feature=de.to(device), lstm_input=en.to(device))
                start = en
            else:
                ago = start[:, 1:, :]  # 除去第一列
            pred.append(out.item())
            tagt.append(tg.item())
    diff = np.abs(np.array(np.power(2, tagt)) - np.array(np.power(2, pred)))
    _MAPE = np.sum(diff)
    print('MAPE: ', _MAPE)
    if _MAPE < MAPE:
        MAPE = _MAPE
        # 保存模型
        torch.save(model, "./Best_transformer.pkl")
        print("The model updated...")
    return MAPE

parser = argparse.ArgumentParser(description='PyTorch Time series forecasting')


parser.add_argument('--optim', type=str, default='adam')
parser.add_argument('--L1Loss', type=bool, default=True)
parser.add_argument('--epochs', type=int, default=10000, help='Number of epochs to train.')
parser.add_argument('--lr', type=float, default=0.005, help='Initial learning rate.')
parser.add_argument('--weight_decay', type=float, default=5e-4, help='Weight decay (L2 loss on parameters).')
parser.add_argument('--hidden', type=int, default=40, help='Number of hidden units.')
parser.add_argument('--nb_heads', type=int, default=8, help='Number of head attentions.')
parser.add_argument('--dropout', type=float, default=0.1, help='Dropout rate (1 - keep probability).')
parser.add_argument('--alpha', type=float, default=0.1, help='Alpha for the leaky_relu.')
parser.add_argument('--num_nodes',type=int,default=15,help='number of nodes/variables')
parser.add_argument('--max_neigh',type=int,default=3,help='k')
parser.add_argument('--node_dim',type=int,default=40,help='dim of nodes')
parser.add_argument('--device',type=str,default='cuda:0',help='which device to use')




args = parser.parse_args()
device = torch.device(args.device)



Epoch = 1000
# Model parameters


# Config
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
print(f"Using device {device}")

model = GAT(nfeat=24,
            nhid=args.hidden,
            nclass= 1,
            dropout=args.dropout,
            nheads=args.nb_heads,
            alpha=args.alpha,
            num_nodes = args.num_nodes,
            max_neigh = args.max_neigh,
            node_dim = args.node_dim,
            device = args.device,
            add_feature_dim = 4,
            feature_size = 14, addition_feature_size=4, output_dim= 1, encoder_length=24,decoder_length=1,lstm_size=128)
model = model.to(device)



# Load transformer with Adam optimizer and MSE loss function

#model = Att_NMT(feature_size = 13, encoder_length=24, add_dimen=3,lstm_size=256, training=True)
criterion = nn.MSELoss(reduce=True)
# criterion = nn.MSELoss()
optimizer = optim.Adam(model.parameters())

# 加载数据
trainData = TrainData()
genTrainData = DataLoader(trainData, batch_size=32, shuffle=True)
testData = TestData()
genTestData = DataLoader(testData, batch_size=1)


sample_nbr = 3


# 训练
MAPE = np.inf
perm = np.random.permutation(range(args.num_nodes))#
for ep in range(Epoch):
    loss_mean = 0.
    model.train()
    for i, (feature, add_feature, tg) in enumerate(genTrainData):
        optimizer.zero_grad()
        feature, add_feature, tg = feature.to(device), add_feature.to(device), tg.to(device)

        # feature = torch.unsqueeze(feature,dim=1)
        #
        feature_gnn = feature.transpose(-2,-1)


        perm = np.random.permutation(range(args.num_nodes))# 随机的产生一些index

        id = perm[ : ]

        id = torch.tensor(id).to(device)

        tx = feature_gnn[:, id.type(torch.long), :]

        out, _ = model(tx, idx=id, add_feature = add_feature, lstm_input=feature)
        m = out.reshape(-1, 1)

        # loss=0

        loss = criterion(m, tg)

        loss.backward()
        # update
        optimizer.step()
        # 打印训练信息
        loss_mean += loss.item()
        if (i + 1) % 10 == 0:
            loss_mean = loss_mean / 10
            print("Training:Epoch[{:0>3}/{:0>3}] Iteration[{:0>3}/{:0>3}] Loss: {:.10f}".format(ep, Epoch, i + 1, len(genTrainData), loss_mean))
            loss_mean = 0.

    # 测试
    if ep > 5:
        MAPE = test(model, genTestData, MAPE, device)
    torch.save(model, "./transformer.pkl")